# DRF
from rest_framework import mixins, filters
from rest_framework.viewsets import GenericViewSet
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status

# Third Party
from django_filters.rest_framework import DjangoFilterBackend
from drf_yasg.utils import swagger_auto_schema
from drf_yasg import openapi

# App
from dynamic_contents import pagination
from .serializers import FormatSerializer, PartSerializer
from .models import Format, Part, DynamicContent
from .serializers import DynamicContentSerializerMixin

# Classes
class BaseGenericViewSet(GenericViewSet):
    filter_backends = [filters.OrderingFilter, filters.SearchFilter, DjangoFilterBackend]
    pagination_class = pagination.DefaultPagination


class FormatViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    BaseGenericViewSet
):
    queryset = Format.objects.all()
    serializer_class = FormatSerializer
    search_fields = ['type', 'content']
    ordering_fields = ['created_at', 'type']


class PartViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    BaseGenericViewSet
):
    queryset = Part.objects.all()
    serializer_class = PartSerializer
    search_fields = ['field', 'content', 'link']
    ordering_fields = ['created_at', 'type']
    filterset_fields = ['instance_id']


class DynamicContentView(APIView):
    @swagger_auto_schema(
        manual_parameters=[
            openapi.Parameter(
                'parts', openapi.IN_QUERY,
                description="Comma-separated list of part IDs",
                type=openapi.TYPE_STRING
            )
        ],
        responses={200: openapi.Response('Dynamic content response')}
    )
    def get(self, request, format_id):
        try:
            format_instance = Format.objects.get(pk=format_id)
            parts_ids = request.query_params.get('parts', '')

            # 쉼표로 분리하여 parts_ids를 리스트로 변환
            parts_ids_list = [int(pid) for pid in parts_ids.split(',') if pid.isdigit()]

            parts_instances = Part.objects.filter(id__in=parts_ids_list)

            # DynamicContent 인스턴스 생성
            dynamic_content = DynamicContent(format_instance, parts_instances)

            # DynamicContent 객체를 사용하여 최종 콘텐츠 생성
            response_data = DynamicContentSerializerMixin(dynamic_content).data
            return Response(response_data)
        except Format.DoesNotExist:
            return Response({"error": "Format not found"}, status=status.HTTP_404_NOT_FOUND)
        except ValueError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)



# Python
import re
from collections import defaultdict

# Django
from django.utils.translation import gettext_lazy as _


def group_parts_by_field(parts):
    grouped_parts = defaultdict(list)
    for part in parts if type(parts) == list else parts.all():
        grouped_parts[part.field].append(part.get_content())
    return grouped_parts


def join_contents(contents):
    # 마지막 요소 전까지는 쉼표로, 마지막 두 요소는 "and"로 연결
    if len(contents) > 1:
        return ', '.join(contents[:-1]) + _(' and ') + contents[-1]
    elif contents:
        return contents[0]
    return ''


def generate_text(format, parts):
    """
    Generates text from a format and parts, joining multiple contents for the same field.
    """
    if not format:
        return ''

    format_string = format.get_content()

    grouped_parts = group_parts_by_field(parts)
    for field, contents in grouped_parts.items():
        joined_contents = join_contents(contents)
        if joined_contents:
            format_string = format_string.replace(f"{{{{{field}}}}}", joined_contents)
    return format_string


def generate_i18n(format, parts):
    if not format:
        return ''

    format_string = format.get_content()

    # grouped_parts에 각 part.field 별로 part.content를 그룹화합니다.
    grouped_parts = defaultdict(list)
    for part in parts if type(parts) == list else parts.all():
        grouped_parts[part.field].append(part)

    # placeholders를 찾아서 각 placeholder에 대한 초기 인덱스 값을 매핑합니다.
    placeholders = re.findall(r'\{\{(\w+)\}\}', format_string)
    placeholder_indices = 0

    for placeholder in placeholders:
        if placeholder in grouped_parts:
            # 해당 placeholder에 대한 parts 리스트를 가져옵니다.
            parts_for_placeholder = grouped_parts[placeholder]

            contents = []
            for i, part in enumerate(parts_for_placeholder):
                # 현재 인덱스 값을 사용합니다.
                contents.append(f'<{placeholder_indices}>{part.get_content()}</{placeholder_indices}>')
                # 다음 placeholder에 대한 인덱스를 업데이트합니다.
                placeholder_indices += 1

            joined_contents = join_contents(contents)
            format_string = re.sub(f"{{{{{placeholder}}}}}", joined_contents, format_string, count=1)

    return format_string


def generate_html(format, parts):
    if not format:
        return ''

    format_string = format.get_content()

    grouped_parts = defaultdict(list)

    for part in parts if type(parts) == list else parts.all():
        link = part.link if hasattr(part, 'link') and part.link else '#'
        html_link = f'<a href="{link}">{part.get_content()}</a>'
        grouped_parts[part.field].append(html_link)

    for field, html_links in grouped_parts.items():
        joined_html_links = join_contents(html_links)
        format_string = format_string.replace(f"{{{{{field}}}}}", joined_html_links)

    return format_string



# Python
import logging
import requests
from typing import Any
from urllib.parse import urljoin

# Django
from django.conf import settings


# Variables
logger = logging.getLogger(__name__)


# Classes
class BaseGateway:
    def __init__(self, base_url: str):
        self.base_url = base_url

    def get_url(self, path: str) -> str:
        return urljoin(self.base_url, path)

    def request(self, method: str, path: str, *args, **kwargs) -> Any:
        try:
            response = requests.request(method, self.get_url(path), *args, **kwargs)
        except requests.RequestException as exc:
            # Write logs for possible request errors
            logger.warning(f"Unexpected exception caught: {exc!s}")
            raise

        # Raise exception for HTTP error status codes such as 400, 404, 500.
        # response.raise_for_status()

        return response.json()


class GatewayV1(BaseGateway):
    def __init__(self):
        super().__init__(base_url=urljoin(settings.GATEWAY1_HOST, f'/api/'))

    def func1(self, field_value1: str, field_value2: str):
        path = 'path1/path2/...'

        body = {
            "field_name1": field_value1,
            "field_name2": field_value2,
        }
        print('[func1] body : ', body)

        return self.request(method="POST", path=path, json=body)


# Instances
package_sdk = GatewayV1()



# vim: set fileencoding=utf-8 :
from django.conf import settings
from django.contrib import admin
from django.utils.safestring import mark_safe

# Third Party
from modeltranslation.admin import TranslationAdmin

# App
from .models import Format, Part


# Main Section
class FormatAdmin(admin.ModelAdmin):
    # 기본 list_display 설정
    list_display = ('id', 'type', 'subtype', 'content', 'created_at', 'updated_at')

    def __init__(self, model, admin_site):
        super().__init__(model, admin_site)
        # modeltranslation으로 생성된 번역 필드를 가져와서 list_display에 추가
        translated_fields = self.get_translated_fields()
        self.list_display = self.list_display + translated_fields

    def get_translated_fields(self):
        # 모델의 모든 필드를 순회하면서 번역 필드 확인
        translated_fields = []
        for field in self.model._meta.get_fields():
            if self.is_translated_field(field.name):
                translated_fields.append(field.name)
        return tuple(translated_fields)

    def is_translated_field(self, field_name):
        # 원래 필드명으로부터 생성된 번역 필드인지 확인
        return '_' in field_name and field_name.split('_')[-1] in dict(settings.LANGUAGES).keys()


class PartAdmin(admin.ModelAdmin):
    list_display = ('id', 'field', 'content', 'link', 'instance_id', 'created_at', 'updated_at')
    search_fields = ('field', 'content', 'instance_id')


class PartInline(admin.TabularInline):
    model = Part
    extra = 1  # 기본적으로 보여줄 빈 인라인 폼의 수


class DynamicContentAdminMixin(admin.ModelAdmin):
    append_list_display = ('text_content', 'i18n_content', 'html_content', 'missing_placeholders',)
    append_list_filter = ('format',)
    append_readonly_fields = ('format', 'parts', 'missing_placeholders')

    def text_content(self, obj):
        return obj.get_text()
    text_content.short_description = 'Text Content'

    def i18n_content(self, obj):
        return obj.get_i18n()
    i18n_content.short_description = 'I18N Content'

    def html_content(self, obj):
        return mark_safe(obj.get_html())
    html_content.short_description = 'HTML Content'
    html_content.allow_tags = True

    def get_list_display(self, request):
        parent_fields = super().get_list_display(request)
        combined = dict.fromkeys(list(parent_fields) + list(self.append_list_display))
        return tuple(combined.keys())

    def get_list_filter(self, request):
        parent_fields = super().get_list_filter(request)
        combined = dict.fromkeys(list(parent_fields) + list(self.append_list_filter))
        return tuple(combined.keys())

    def get_readonly_fields(self, request, obj=None):
        parent_fields = super().get_readonly_fields(request, obj)
        combined = dict.fromkeys(list(parent_fields) + list(self.append_readonly_fields))
        return tuple(combined.keys())


admin.site.register(Format, FormatAdmin)
admin.site.register(Part, PartAdmin)



# Python
import re
import json
from collections import Counter

# Django
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils.translation import get_language
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder

# App
from .utils import generate_text, generate_html, generate_i18n

# Class Section
class BaseModel(models.Model):

    # Dates
    created_at = models.DateTimeField(_('Created At'), auto_now_add=True, null=True)
    updated_at = models.DateTimeField(_('Updated At'), auto_now=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.id)

    def get_content(self):
        """
        현재 활성화된 언어에 맞는 content 값을 반환합니다.
        """
        current_language = get_language()
        # content 필드에 대한 현재 언어 버전의 속성명을 생성합니다.
        content_field_name = f"content_{current_language.replace('-', '_')}"
        # 현재 언어 버전의 content 값을 반환하되, 없다면 기본 content 값을 사용합니다.
        if content := getattr(self, content_field_name, None):
            return content
        else:
            return self.content


class FormatManager(models.Manager):
    def update_format_if_needed(self, format, defaults):
        updated = False
        for key, value in defaults.items():
            if not getattr(format, key):
                setattr(format, key, value)
                updated = True
        if updated:
            format.save()


# Format
class Format(BaseModel):

    type = models.CharField(_('Type'), max_length=100)
    subtype = models.CharField(_('Sub Type'), max_length=100)
    content = models.TextField(_('Content'))  # "{user}가 {post}를 좋아요합니다."
    _placeholders = models.TextField(_('Placeholders'), blank=True, null=True)

    objects = FormatManager()

    class Meta:
        verbose_name = 'format'
        verbose_name_plural = 'formats'
        ordering = ['-created_at']

    def __str__(self):
        return '{}({}) {}'.format(self.__class__.__name__, self.id, self.get_content())

    def get_placeholders(self):
        """
        'placeholders' 필드를 배열로 변환하여 반환합니다.
        """
        if self._placeholders:
            return self._placeholders.split(',')
        return []

    def save(self, *args, **kwargs):
        # type과 subtype 필드를 대문자로 변환하고, _ 외의 특수문자 제거
        self.type = self.process_type_field(self.type)
        self.subtype = self.process_type_field(self.subtype)
        self._placeholders = self.extract_placeholders(self.content)

        placeholders = self.get_placeholders()
        placeholder_counts = Counter(placeholders)
        for placeholder, count in placeholder_counts.items():
            if count > 1:
                raise ValidationError(f'Placeholder "{{{placeholder}}}" used multiple times in content.')

        return super(Format, self).save(*args, **kwargs)

    @staticmethod
    def process_type_field(field_value):
        """
        필드 값을 대문자로 변환하고, _ 외의 특수문자를 제거합니다.
        """
        if field_value:
            return re.sub(r'[^A-Z_/]', '', field_value.upper())
        return field_value

    @staticmethod
    def extract_placeholders(content):
        """
        주어진 content에서 '{{}}' 형식의 텍스트를 추출합니다.
        """
        if not content:
            return ''

        placeholders = re.findall(r'\{\{(\w+)\}\}', content)
        return ','.join(placeholders)


# Part
class Part(BaseModel):
    field = models.TextField(_('Field'), null=True, blank=True)  # user
    content = models.TextField(_('Content'), null=True, blank=True)  # 김선욱
    link = models.URLField(_('Link'), null=True, blank=True)  # https://runners.im/sun
    instance_id = models.TextField(_('Instance ID'), null=True, blank=True)  # 1

    class Meta:
        verbose_name = 'part'
        verbose_name_plural = 'parts'
        ordering = ['-created_at']

    def __str__(self):
        return '{}({}) {}'.format(self.__class__.__name__, self.id, self.content)

    def save(self, *args, **kwargs):
        super(Part, self).save(*args, **kwargs)


# Dynamic Content
class DynamicContentManagerMixin:

    def create_dynamic_content(self, format, parts):
        """
        Create a new DynamicContent object with the given format and parts.

        :param format: The Format object for the DynamicContent.
        :param parts: List of Part objects.
        :return: The created DynamicContent object.
        """
        # DynamicContent 객체 생성
        dynamic_content = self.create(format=format)

        # Part 객체들 연결
        for part in parts:
            dynamic_content.parts.add(part)

        return dynamic_content

    def update_dynamic_content(self, dynamic_content, format, parts):
        """
        Update the format and parts of the given DynamicContent object.

        :param dynamic_content: The DynamicContent object to update.
        :param format: The Format object for the DynamicContent.
        :param parts: List of Part objects.
        :return: The updated DynamicContent object.
        """
        # 기존 Part 객체들 삭제
        dynamic_content.parts.clear()

        # Part 객체들 연결
        for part in parts:
            dynamic_content.parts.add(part)

        # Format 업데이트
        dynamic_content.format = format

        # DynamicContent 저장
        dynamic_content.save()

        return dynamic_content


class DynamicContentModelMixin(models.Model):

    format = models.ForeignKey(Format, on_delete=models.SET_NULL, null=True, blank=True)
    parts = models.ManyToManyField(Part, blank=True)
    missing_placeholders = models.TextField(_('Missing Placeholders'), blank=True, null=True)

    class Meta:
        abstract = True

    def get_missing_placeholders(self):
        """
        이 메서드는 Format의 placeholders와 연결된 Parts가 모두 존재하는지 확인합니다.
        누락된 placeholders가 있다면 해당 placeholders를 리스트로 반환합니다.
        """
        if not self.format:
            return []  # Format이 설정되지 않은 경우 빈 리스트 반환

        placeholders = self.format.get_placeholders()
        parts_fields = [part.field for part in self.parts.all()]
        missing_placeholders = list(set(placeholders) - set(parts_fields))

        return missing_placeholders

    def get_text(self):
        # generate_text 함수를 사용하여 텍스트 생성
        return generate_text(self.format, self.parts.all())

    def get_i18n(self):
        # generate_i18n 함수를 사용하여 국제화된 텍스트 생성
        return generate_i18n(self.format, self.parts.all())

    def get_html(self):
        # generate_html 함수를 사용하여 HTML 콘텐츠 생성
        return generate_html(self.format, self.parts.all())

    def save(self, *args, **kwargs):
        if self.id:
            # Update the missing_placeholders field before saving
            missing = self.get_missing_placeholders()
            self.missing_placeholders = json.dumps(missing, cls=DjangoJSONEncoder)

        super(DynamicContentModelMixin, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        DynamicContent를 삭제하기 전에 관련된 모든 파트를 삭제합니다.
        """
        self.parts.all().delete()
        super(DynamicContentModelMixin, self).delete(*args, **kwargs)


class DynamicContent:
    def __init__(self, format, parts):
        """
        Initializes a utility object to work with dynamic content.

        :param format_instance: An instance or mock of the Format model.
        :param parts_queryset: A list or mock queryset of Part instances.
        """
        self.format = format
        self.parts = parts

    def get_missing_placeholders(self):
        if not self.format:
            return []

        placeholders = self.format.get_placeholders()
        parts_fields = [part.field for part in self.parts]
        missing_placeholders = list(set(placeholders) - set(parts_fields))

        return missing_placeholders

    def get_text(self):
        # Assuming generate_text is a standalone function that processes the format and parts
        return generate_text(self.format, self.parts)

    def get_i18n(self):
        # Assuming generate_i18n is a standalone function
        return generate_i18n(self.format, self.parts)

    def get_html(self):
        # Assuming generate_html is a standalone function
        return generate_html(self.format, self.parts)




import unittest

from dynamic_contents.utils import generate_text, generate_i18n, generate_html


class MockFormat:
    def __init__(self, content):
        self.content = content

    def get_content(self):
        return self.content

class MockPart:
    def __init__(self, field, content, link=None):
        self.field = field
        self.content = content
        self.link = link

    def get_content(self):
        return self.content

class TestDynamicContentUtils(unittest.TestCase):
    def setUp(self):
        self.format = MockFormat("Hello, {{user}}! Your post {{post}} was liked by {{user_other}} for {{user_the_other}}.")
        self.parts = [
            MockPart("user", "Alice"),
            MockPart("post", "How to learn Python"),
            MockPart("user_other", "Leo", link="http://example.com"),
            MockPart("user_the_other", "Rosie", link="http://example.com")
        ]

    def test_generate_text(self):
        expected_text = "Hello, Alice! Your post How to learn Python was liked by Leo for Rosie."
        print(generate_text(self.format, self.parts))
        self.assertEqual(generate_text(self.format, self.parts), expected_text)

    def test_generate_i18n(self):
        expected_i18n = "Hello, <0>Alice</0>! Your post <1>How to learn Python</1> was liked by <2>Leo</2> for <3>Rosie</3>."
        print(generate_i18n(self.format, self.parts))
        self.assertEqual(generate_i18n(self.format, self.parts), expected_i18n)

    def test_generate_html(self):
        expected_html = 'Hello, <a href="#">Alice</a>! Your post <a href="#">How to learn Python</a> was liked by <a href="http://example.com">Leo</a> for <a href="http://example.com">Rosie</a>.'
        print(generate_html(self.format, self.parts))
        self.assertEqual(generate_html(self.format, self.parts), expected_html)

if __name__ == '__main__':
    unittest.main()
